
/*
 * $Id: OLog.h,v 1.7.6.3 1999/11/02 16:32:01 rousskov Exp $
 *
 *
 * Web Polygraph                        http://polygraph.ircache.net/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1999 by Alex Rousskov and the
 *  University of California, San Diego (UCSD); all rights reserved.
 *
 *  You may copy and distribute VERBATIM copies of this product. 
 *  You may modify your copy for your own use. You must NOT
 *  distribute modified copies. You must NOT publish benchmarking
 *  results based on modified versions.
 *
 *  If you publish any materials derived from or related to Web
 *  Polygraph, you must give a credit reference to the IRCache group.
 *
 *  Web Polygraph is distributed without any warranty, even implied.
 *
 *  Please see the COPYRIGHT file for full details.
 *
 */

#ifndef O_LOG_CLASS_H
#define O_LOG_CLASS_H

#include "Time.h"
#include "Size.h"
#include "NetDouble.h"
#include "BigSize.h"
#include "String.h"
#include "NetAddr.h"
#include "Array.h"

// some environments do not know better than #define these
#ifdef getc
#undef getc
#endif
#ifdef putc
#undef putc
#endif

// buffered binary output log
class OLog {
	public:
		typedef void (Manip)(class OLog &);

	public:
		OLog();
		virtual ~OLog();

		Size capacity() const { return theCapacity; }
		Size size() const { return theSize; }

		void stream(const String &aName, int aFD);
		void resize(Size aCap);

		void putc(char c) { put(&c, 1); }
		void putb(bool b) { putc((char)(b ? 1 : 0)); }
		void puti(int x) { x = htonl(x); put(&x, SizeOf(x)); }
		void puti(const int *x, int count);
		void puts(const char *s, Size size) { puti(size); if (size) put(s, size); }

		virtual void begEntry(int tag);
		void endEntry();

		void flush(Size maxSize); // flushes up-to max size
		void flush() { flush(theSize); }
		void close(); // flushes if needed

	protected:
		inline void put(const void *buf, Size size);
		void overflow(const void *buf, Size size);

		virtual void putHeader();
		virtual void putTrailer();

	protected:
		int theFD;
		String theName;

		char *theBuf;
		Size theCapacity; // buffer space allocated
		Size theSize;     // buffer space used
		Size thePos;      // global offset

		char *theEntry;	  // start of the current entry
		int theEntryTag;  // tag saved until endEntry()

		Array<int> theDir; // directory (one entry per tag)
};


/* manipulators to begin and end logging of a named entry */

struct bege {
	bege(int t): tag(t) {}
	int tag;
};

inline void ende(OLog &log) { log.endEntry(); }


/* logging of common types */

inline
OLog &operator <<(OLog &ol, char c) {
	ol.putc(c);
	return ol;
}

inline
OLog &operator <<(OLog &ol, bool b) {
	ol.putb(b);
	return ol;
}

inline
OLog &operator <<(OLog &ol, int x) {
	ol.puti(x);
	return ol;
}

inline 
OLog &operator <<(OLog &ol, double x) { 
	NetDouble nd(x);
	return ol << nd.mnt << nd.exp;
}

inline
OLog &operator <<(OLog &ol, const char *s) {
	ol.puts(s, strlen(s)+1);
	return ol;
}

inline
OLog &operator <<(OLog &ol, const String &s) {
	if (s)
		ol.puts(s.str(), s.len()+1);
	else
		ol.puts(0, 0);
	return ol;
}

inline
OLog &operator <<(OLog &ol, const Time &t) {
	ol.puti(t.tv_sec);
	ol.puti(t.tv_usec);
	return ol;
}

inline
OLog &operator <<(OLog &ol, const BigSize &bs) {
	return ol << bs.theAcc << bs.theCnt;
}

inline
OLog &operator <<(OLog &ol, const NetAddr &a) {
	return ol << a.addrA() << (int)a.addrN().s_addr << a.port();
}

inline
OLog &operator <<(OLog &ol, const bege &b) {
	ol.begEntry(b.tag);
	return ol;
}

inline
OLog &operator <<(OLog &ol, OLog::Manip m) {
	m(ol);
	return ol;
}

template <class Item>
inline
OLog &operator <<(OLog &ol, const Array<Item> &a) {
	ol.puti(a.count());
	for (int i = 0; i < a.count(); ++i)
		ol << a[i];
	return ol;
}


/* inlined methods */

inline
void OLog::put(const void *buf, Size size) {
	if (size > 0 && theFD >= 0) {
		// just append if fits
		if (theSize + size <= theCapacity) {
			memcpy(theBuf + theSize, buf, size);
			theSize += size;
			thePos += size;
		} else {
			overflow(buf, size); // always uses put() recursively
		}
	}
}

#endif
