
/*
 * $Id: Array.h,v 1.7.6.5 1999/11/29 03:12:40 rousskov Exp $
 *
 *
 * Web Polygraph                        http://polygraph.ircache.net/
 * ------------------------------------------------------------------
 *
 *  Web Polygraph is copyrighted (C) 1999 by Alex Rousskov and the
 *  University of California, San Diego (UCSD); all rights reserved.
 *
 *  You may copy and distribute VERBATIM copies of this product. 
 *  You may modify your copy for your own use. You must NOT
 *  distribute modified copies. You must NOT publish benchmarking
 *  results based on modified versions.
 *
 *  If you publish any materials derived from or related to Web
 *  Polygraph, you must give a credit reference to the IRCache group.
 *
 *  Web Polygraph is distributed without any warranty, even implied.
 *
 *  Please see the COPYRIGHT file for full details.
 *
 */

#ifndef X_ARRAY_H
#define X_ARRAY_H

#include <string.h>

#include "Assert.h"

// Array is an array of <Item> items with an unlimited capacity
// note: for optimization purposes, we mem-copy Items if needed
template <class Item>
class Array {
	public:
		Array(int aCapacity = 0): theItems(0), theCapacity(aCapacity), theCount(0) { theItems = new Item[aCapacity]; }
		~Array() { delete[] theItems; }

		bool full() const { return theCount >= theCapacity; }
		bool empty() const { return !theCount; }
		int capacity() const { return theCapacity; }
		int count() const { return theCount; }
		int size() const { return theCapacity*sizeof(Item); }

		Item &item(int idx) { return theItems[idx]; }
		const Item &item(int idx) const { return theItems[idx]; }
		Item &operator [](int idx) { return item(idx); }
		const Item &operator [](int idx) const { return item(idx); }
		Item &last(int off = 0) { return theItems[theCount-1-off]; }
		const Item &last(int off = 0) const { return theItems[theCount-1-off]; }

		void reset() { theCount = 0; }
		void stretch(int n) { if (n > theCapacity) grow(n); }
		void count(int aCount) { Assert(aCount <= theCapacity); theCount = aCount; }

		inline void append(const Item &i);
		inline Item shift() { Item i(theItems[0]); shift(1); return i; }
		inline void shift(int n);
		void push(const Item &i) { append(i); }
		Array<Item> &operator <<(const Item &i) { append(i); return *this; }
		void pop(int n) { Assert(theCount >= n); theCount -= n; }
		Item pop() { Assert(theCount > 0); return theItems[--theCount]; }
		Array<Item> &operator >>(Item &i) { i = pop(); return *this; }
		void put(const Item &i, int idx) { stretch(idx+1); theItems[idx] = i; if (theCount <= idx) theCount = idx+1; }
		void swap(int i, int j) { const Item h = theItems[i]; theItems[i] = theItems[j]; theItems[j] = h; }

		void memset(int v, int off = 0) { ::memset(theItems+off, v, (theCapacity-off)*sizeof(Item)); }
		void memmove(int off, const Item *from, int cnt) { ::memmove(theItems+off, from, cnt*sizeof(Item)); }
		const Item *items() const { return theItems; }
		Item *items() { return theItems; }

		inline Array &operator =(const Array &arr);

	protected:
		Array(const Array &) {} // no support for silent copying

		void grow() { grow(theCapacity <= 0 ? 16 : theCapacity*2); }
		inline void grow(int newCap);

	protected:
		Item *theItems;
		int theCapacity;
		int theCount;
};

/* searchable array */
template <class Item>
class SchArray: public Array<Item> {
	public:
		SchArray(int aCapacity = 0): Array<Item>(aCapacity) {}
		inline bool find(const Item &i, int &idx) const; // first matching item
		inline bool findOther(const Item &i, int &idx) const; // first different item
};


/* implementaion of in-lined methods */

template <class Item>
inline
void Array<Item>::append(const Item &item) {
    if (theCount >= theCapacity)
		grow();
    theItems[theCount++] = item;
}

// efficient, but not very generic
template <class Item>
inline
void Array<Item>::grow(int newCap) {
	Assert(newCap >= theCount);
	Item *oldItems = theItems;
	theItems = new Item[newCap];
	if (oldItems && theCount)
		memcpy(theItems, oldItems, theCount*sizeof(Item));
	delete[] oldItems;
	theCapacity = newCap;
	memset(0, theCount);
}

// discards first n items
template <class Item>
inline
void Array<Item>::shift(int n) {
	
	if (theCount > n) {
		theCount -= n;
		memmove(0, theItems + n, theCount);
	} else
		theCount = 0;
	memset(0, theCount);
}

template <class Item>
inline
Array<Item> &Array<Item>::operator =(const Array<Item> &arr) {
	stretch(arr.count());
	for (int idx = 0; idx < arr.count(); ++idx)
		append(arr[idx]);
	return *this;
}

template <class Item>
inline
bool SchArray<Item>::find(const Item &i, int &idx) const {
	for (idx = 0; idx < theCount; ++idx)
		if (item(idx) == i)
			return true;
	return false;
}

template <class Item>
inline
bool SchArray<Item>::findOther(const Item &i, int &idx) const {
	for (idx = 0; idx < theCount; ++idx)
		if (item(idx) != i)
			return true;
	return false;
}


#endif
